from __future__ import annotations

import itertools
from collections.abc import ItemsView, Iterable, Iterator, KeysView, Mapping, ValuesView
from typing import TypeVar, overload

from mlc._cython import Ptr
from mlc.dataclasses import c_class

from .object import Object

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


@c_class("object.Dict", init=False)
class Dict(Object, Mapping[K, V]):
    capacity: int
    size: int
    data: Ptr

    def __init__(
        self,
        _source: Iterable[tuple[K, V]] | Mapping[K, V] | None = None,
        **kwargs: V,
    ) -> None:
        if isinstance(_source, Mapping):
            py_args = tuple(_source.items())
        elif isinstance(_source, Iterable):
            py_args = tuple(_source)  # type: ignore[arg-type]
        elif _source is None:
            py_args = ()
        else:
            raise TypeError(f"expected mapping or iterable, got {type(_source).__name__}")
        if kwargs:
            py_args += tuple(kwargs.items())
        self._mlc_init(*itertools.chain.from_iterable(py_args))

    def __len__(self) -> int:
        return self.size

    def __iter__(self) -> Iterator[K]:
        return self._keys_iterator()

    def _keys_iterator(self) -> Iterator[K]:
        cap = self.capacity
        i = -1
        while (i := Dict._C("__iter_advance__", self, i)) < cap:
            yield Dict._C("__iter_get_key__", self, i)

    def _values_iterator(self) -> Iterator[V]:
        cap = self.capacity
        i = -1
        while (i := Dict._C("__iter_advance__", self, i)) < cap:
            yield Dict._C("__iter_get_value__", self, i)

    def _items_iterator(self) -> Iterator[tuple[K, V]]:
        cap = self.capacity
        i = -1
        while (i := Dict._C("__iter_advance__", self, i)) < cap:
            yield Dict._C("__iter_get_key__", self, i), Dict._C("__iter_get_value__", self, i)

    def keys(self) -> KeysView[K]:
        return _DictKeysView(self)

    def values(self) -> ValuesView[V]:
        return _DictValuesView(self)

    def items(self) -> ItemsView[K, V]:
        return _DictItemsView(self)

    def __getitem__(self, key: K) -> V:
        return Dict._C("__getitem__", self, key)

    # Additional methods required by the Mapping ABC
    def __contains__(self, key: object) -> bool:
        try:
            self[key]  # type: ignore
            return True
        except KeyError:
            return False

    @overload
    def get(self, key: K) -> V | None: ...

    @overload
    def get(self, key: K, default: V | T) -> V | T: ...

    def get(self, key: K, default: V | T | None = None) -> V | T | None:
        try:
            return self[key]
        except KeyError:
            return default


class _DictKeysView(KeysView[K]):
    def __init__(self, mapping: Dict[K, V]) -> None:
        self._mapping = mapping

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[K]:
        return self._mapping._keys_iterator()

    def __contains__(self, key: object) -> bool:
        return key in self._mapping


class _DictValuesView(ValuesView[V]):
    def __init__(self, mapping: Dict[K, V]) -> None:
        self._mapping = mapping

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[V]:
        return self._mapping._values_iterator()

    def __contains__(self, value: object) -> bool:
        return any(v == value for v in self)


class _DictItemsView(ItemsView[K, V]):
    def __init__(self, mapping: Dict[K, V]) -> None:
        self._mapping = mapping

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[tuple[K, V]]:
        return self._mapping._items_iterator()

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, tuple) or len(item) != 2:
            return False
        key, value = item
        try:
            v = self._mapping[key]
        except KeyError:
            return False
        else:
            return v == value
